'''Models for basic Data Clustestering and Analysis'''


import pandas as pd
from sklearn.tree import plot_tree
from sklearn.neighbors import KNeighborsClassifier
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
from scipy import cluster
from scipy.spatial.distance import pdist
import numpy as np
from scipy.cluster.hierarchy import cut_tree
from .result_evaluate import confuseResults

def PCA_Decompsition(
    data
):
    pcs = PCA(2)
    df = pcs.fit_transform(data)
    return df

def KMeans_Cluster_Plot(
    df_dec,
    n_class = 2
):
    kmeans = KMeans(n_clusters = 2)
    label = kmeans.fit_predict(df_dec)
    u_labels = np.unique(label)
    centroids = kmeans.cluster_centers_
    for i in u_labels:
        plt.scatter(df_dec[label == i,0], df_dec[label == i,1],label = i)
    plt.scatter(centroids[:,0], centroids[:,1], s = 80)
    plt.legend()
    plt.show()

def KNeigbourClassResults(
    X,
    y
):
    neigh = KNeighborsClassifier(n_neighbors = 2)
    neigh.fit(X,y)
    confuseResults(y,neigh.predict(X))

def AggloClust(
    X,
    y,
    data
):
    Z = cluster.hierarchy.ward(pdist(X))
    cut = cut_tree(Z, n_clusters=2)
    labels = list([i[0] for i in cut])
    fig, axes = plt.subplot(nrow = 1, ncols = 1, figsize = (4,4), dpi = 300)
    plot_tree(cut, feature_names=labels, filled=True, class_names=True)




'''Model evaluations'''

from sklearn.metrics import confusion_matrix


def confuseResults(
    yhat,
    ypred
):
    cm = confusion_matrix(y_true=yhat,y_pred=ypred)
    print('Confusion Matrix: \n',cm)
    total = sum(sum(cm))
    print('From Confusion Matric we calculate addtional parameters')
    print('Accuracy : ',(cm[0,0]+cm[1,1])/total)
    print('Sensitivity: ', cm[0,0]/(cm[0,0]+cm[0,1]))
    print('Specificity : ',cm[1,1]/(cm[1,0]+cm[1,1]))



'''Preprocessing the data to improve the perfomence of classifier'''

'''
Inorder to imporve the perfomence we are going to use 4 resampling methods
* Simple Random over-sampling
* Cluster based over-sampling
* Neighbourhood Cleaning Rule(NCR)
* SPIDER

Apart from these we are also going to implement and use SMOTE(Synthetic Minority Over Sampling)
'''

from collections import Counter
import numpy as np
from sklearn import cluster
from imblearn.over_sampling import SMOTE
from sklearn.cluster import KMeans
from clover.over_sampling import ClusterOverSampler
from imblearn.over_sampling import RandomOverSampler
from imblearn.under_sampling import NeighbourhoodCleaningRule
from multi_imbalance.resampling.spider import SPIDER3
from sklearn.decomposition import PCA 
from collections import Counter

def extract_characterstics(
    X, 
    y
):
    n_samples, n_features = X.shape
    count_y = Counter(y)
    (maj_label, n_samples_maj), (min_label, n_samples_min) = count_y.most_common()
    ir = n_samples_maj / n_samples_min
    return n_samples, n_features, maj_label, n_samples_maj, min_label, n_samples_min, ir

def print_characterstics(
    X,
    y
):
    n_samples, n_features, maj_label, n_samples_maj, min_label, n_samples_min, ir = extract_characterstics(X, y)
    print(
        f'Number of samples: {n_samples}',
        f'Number of features: {n_features}',
        f'Majority class label: {maj_label}',
        f'Number of majority class samples: {n_samples_maj}',
        f'Minority class label: {min_label}',
        f'Number of Minority class sample: {n_samples_min}',
        f'Imbalance Ratio : {ir:.1f}',
        sep='\n'
    )

rnd_seed = 12

def RandomOverSampling(
    X,
    y
):
    ros = RandomOverSampler(random_state = 42)
    X_res, y_res = ros.fit_resample(X, y)
    print("Original Data Characterstics")
    print_characterstics(X,y)
    print("Data Characterstics after random Over Sampling")
    print_characterstics(X_res,y_res)
    return (X_res, y_res)


def ClusterBasedOverSampling(
    X,
    y
):
    smote = SMOTE(random_state = rnd_seed + 1)
    kmeans = KMeans(n_clusters=2, random_state=rnd_seed+5)
    kmeans_smote = ClusterOverSampler(oversampler = smote, clusterer = kmeans)
    X_res, y_res= kmeans_smote.fit_resample(X,y)
    print("Original Data Characterstics")
    print_characterstics(X,y)
    print("Data Characterstics after Cluster Based Over Sampling")
    print_characterstics(X_res,y_res)
    return (X_res, y_res)

def NCR(
    X,
    y
):
    ncr = NeighbourhoodCleaningRule()
    X_res, y_res = ncr.fit_resample(X, y)
    print("Original Data Characterstics")
    print_characterstics(X,y)
    print("Data Characterstics after NCR")
    print_characterstics(X_res,y_res)
    return (X_res, y_res)

def SPIDER(
    X,
    y
):
    cost = np.random.rand(4).reshape((2,2))
    for i in range(2):
        cost[i][i] = 0
    n_samples, n_features, maj_label, n_samples_maj, min_label, n_samples_min, ir = extract_characterstics(X,y)
    maj_int_min = {
        'maj':[maj_label],
        'int':[],
        'min':[min_label]
    }
    clf = SPIDER3(k=1,maj_int_min=maj_int_min,cost = cost)
    X_res, y_res = clf.fit_resample(X, y)
    print("Original Data Characterstics")
    print_characterstics(X,y)
    print("Data Characterstics after SPIDER")
    print_characterstics(X_res,y_res)
    return (X_res, y_res)



'''Simple Cluster plot using MeanShift and Bandwith Estimation'''

import numpy as np
import matplotlib.pyplot as plt
from sklearn.cluster import MeanShift, estimate_bandwidth
from itertools import cycle

def cluster_plot(X,n_clus):
    bandwidth = estimate_bandwidth(X, quantile = 0.1)
    ms = MeanShift(bandwidth = bandwidth, bin_seeding = True)
    ms.fit(X)
    labels = ms.labels_
    cluster_centers = ms.cluster_centers_
    labels_unique = np.unique(labels)
    n_clusters_ = n_clus
    print("Number of estimated clusters : %d " % n_clusters_ )
    plt.figure()
    plt.clf()
    colors = cycle('bgrcmykbgrcmy"kbgrcmykbgrcmyk')
    for k,col in zip(range(n_clusters_),colors):
        my_members = labels == k
        cluster_center = cluster_centers[k]
        plt.plot(X[my_members,0], X[my_members,1], col + '.')
        plt.plot(cluster_center[0], cluster_center[1], 'o', markerfacecolor = col, markeredgecolor = 'k', markersize = 14)
    plt.title('Estimated number of clusters: %d ' % n_clusters_)
    plt.show()


'''Data generation for simulation of class imbalance'''

import numpy as np
import numbers
import pandas as pd
from collections.abc import Iterable
from sklearn.utils import check_array
import unittest

def make_blobs(
    n_samples=100,
    n_features=2,
    *,
    centers=None,
    cluster_std=1.0,
    center_box=(-10.0, 10.0),
    shuffle=True,
    random_state=None,
    return_centers=False,
):
    generator = check_random_state(random_state)
    if isinstance(n_samples, numbers.Integral):

        if centers is None:
            centers = 3

        if isinstance(centers, numbers.Integral):
            n_centers = centers
            centers = generator.uniform(
                center_box[0], center_box[1], size=(n_centers, n_features)
            )

        else:
            centers = check_array(centers)
            n_features = centers.shape[1]
            n_centers = centers.shape[0]

    else:

        n_centers = len(n_samples)
        if centers is None:
            centers = generator.uniform(
                center_box[0], center_box[1], size=(n_centers, n_features)
            )
        try:
            assert len(centers) == n_centers
        except TypeError as e:
            raise ValueError(
                "Parameter `centers` must be array-like. Got {!r} instead".format(
                    centers
                )
            ) from e
        except AssertionError as e:
            raise ValueError(
                "Length of `n_samples` not consistent with number of "
                f"centers. Got n_samples = {n_samples} and centers = {centers}"
            ) from e
        else:
            centers = check_array(centers)
            n_features = centers.shape[1]

    if hasattr(cluster_std, "__len__") and len(cluster_std) != n_centers:
        raise ValueError(
            "Length of `clusters_std` not consistent with "
            "number of centers. Got centers = {} "
            "and cluster_std = {}".format(centers, cluster_std)
        )

    if isinstance(cluster_std, numbers.Real):
        cluster_std = np.full(len(centers), cluster_std)

    X = []
    y = []

    if isinstance(n_samples, Iterable):
        n_samples_per_center = n_samples
    else:
        n_samples_per_center = [int(n_samples // n_centers)] * n_centers

        for i in range(n_samples % n_centers):
            n_samples_per_center[i] += 1

    for i, (n, std) in enumerate(zip(n_samples_per_center, cluster_std)):
        X.append(generator.normal(loc=centers[i], scale=std, size=(n, n_features)))
        y += [i] * n

    X = np.concatenate(X)
    y = np.array(y)

    if shuffle:
        total_n_samples = np.sum(n_samples)
        indices = np.arange(total_n_samples)
        generator.shuffle(indices)
        X = X[indices]
        y = y[indices]

    if return_centers:
        return X, y, centers
    else:
        return X, y



def check_random_state(seed):

    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, numbers.Integral):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(
        "%r cannot be used to seed a numpy.random.RandomState instance" % seed
    )

def center_gen(
    n_samp = 300,
    x_o = 10,
    y_o =10 ,
    rs = 2,
    item = 5,
    lisRe = False
):
    
    
    n_samples = n_samp
    n_bins = 2
    x_0 = x_o
    y_0 = y_o
    r = rs
    items = item
    centers = []
    for i in range(items):
        x = x_0 + r*np.cos(2*np.pi * i/items)
        y = x_0 + r*np.sin(2*np.pi * i/items)
        centers.append((x,y))
    if lisRe:
        return centers
    else:
        return np.array(centers)


def gen_clover5(
    n_samples = 400,
    ratio = 0.5
):
    centers = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = 0.5,item = 5)
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.1, centers = centers, shuffle=True, random_state=42)
    return (X,np.random.binomial(n=1,p=0.5,size=[n_samples]))

def gen_colver3(
    n_samples = 400,
    ratio = 0.5
):
    centers = center_gen(n_samp = 400,x_o = 10,y_o =10 ,rs = 1,item = 3)
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.6, centers = centers, shuffle=True, random_state=42)
    return (X,np.random.binomial(n=1,p=ratio,size=[n_samples]))

def gen_pawdata02b(
    n_samples = 400,
    ratio = 0.5
):
    centers = center_gen(n_samp = 400,x_o = 10,y_o =10 ,rs = 1,item = 3,lisRe = True)
    centers += center_gen(n_samp = 1000,x_o = 12,y_o =12 ,rs = 2,item = 2,lisRe = True)
    centers = np.array(centers)
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.5, centers = centers, shuffle=True, random_state=42)
    return (X,np.random.binomial(n=1,p=ratio,size=[n_samples]))

def gen_pawdata02a(
    n_samples = 400,
    ratio = 0.5
):
    centers = [(10,10),(10,15),(10,5)]
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.5, centers = centers, shuffle=True, random_state=42)
    return (X,np.random.binomial(n=1,p=ratio,size=[n_samples])) 

def gen_subclass3(
    n_samples = 400,
    ratio= 0.5
):
    centers = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = 0.5,item = 6)
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.1, centers = centers, shuffle=True, random_state=42)
    return (X,np.random.binomial(n=1,p=ratio,size=[n_samples])) 

def gen_subclass5(
    n_samples = 400,
    ratio = 0.5
):
    centers = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = 0.5,item = 10)
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.1, centers = centers, shuffle=True, random_state=42)
    return (X,np.random.binomial(n=1,p=ratio,size=[n_samples]))

def arrRatioMaintain1(X,ratio):
    X = np.random.binomial(n=1,p=ratio,size=[len(X)])
    return X
def arrRatioMaintain2(X,ratio):
    X = np.random.binomial(n=1,p=(1-ratio),size=[len(X)])
    return X
def constadd(X,alt):
    if ((alt >= 1) or (alt <= -1)):
        raise ValueError(
            "The Rare Example Generator does not work well with value"
            "of increase/decrease greater than 1 or less than -1. Got Value = {} "
           .format(alt)
        )
    for i in range(0,len(X)):
        X[i] += alt
    return X

    

def gen_overlapping(
    n_subclus = 3,
    n_samples = 400,
    ratio = 0.5
):
    centers1 = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = 0.5,item = n_subclus)
    X1,_  = make_blobs(n_samples = n_samples//2, n_features = 2, cluster_std = 0.1, centers = centers1, shuffle=True, random_state=42)
    centers2 = center_gen(n_samp = 300//2,x_o = 10,y_o =10 ,rs = 0.5,item = n_subclus)
    X2,_  = make_blobs(n_samples = n_samples//2, n_features = 2, cluster_std = 0.1, centers = centers2, shuffle=True, random_state=42)
    Y1 = np.random.binomial(n=1,p=ratio,size=[n_samples//2])
    Y2 = np.random.binomial(n=1,p=ratio,size=[n_samples//2])
    Y1 = np.concatenate((Y1[:(n_samples//2)],arrRatioMaintain1(Y1[(n_samples//2)::],ratio)),axis = 0)
    Y2 = np.concatenate((Y2[:(n_samples//2)],arrRatioMaintain2(Y1[(n_samples//2)::],ratio)), axis = 0)
    X = np.concatenate((X1,X2), axis = 0 )
    y = np.concatenate((Y1,Y2), axis = 0 )
    return (X,y)

def gen_subclus(
    n_samples = 400,
    ratio = 0.5,
    n_clus = 3
):
    centers = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = 0.5,item = n_clus*2)
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.1, centers = centers, shuffle=True, random_state=42)
    return (X,np.random.binomial(n=1,p=ratio,size=[n_samples]))

def gen_rareborder(
    n_clus = 3,
    n_samples = 400,
    ratio = 0.5,
    rs = 0.5,
    val = 1
):
    centers = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = rs,item = n_clus*2)
    X,_  = make_blobs(n_samples = n_samples, n_features = 2, cluster_std = 0.1, centers = centers, shuffle=True, random_state=42)
    Y = np.random.binomial(n=1,p=ratio,size=[n_samples])
    y = np.concatenate((Y[:(n_samples//2)],constadd(Y[(n_samples//2)::],val)),axis = 0)
    return (X,y)

def gen_overlappingRareBorder(
    n_subclus = 3,
    n_samples = 400,
    ratio = 0.5,
    val = 0.99
):
    centers1 = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = 0.5,item = n_subclus)
    X1,_  = make_blobs(n_samples = n_samples//2, n_features = 2, cluster_std = 0.1, centers = centers1, shuffle=True, random_state=42)
    centers2 = center_gen(n_samp = 300,x_o = 10,y_o =10 ,rs = 0.5,item = n_subclus)
    X2,_  = make_blobs(n_samples = n_samples//2, n_features = 2, cluster_std = 0.1, centers = centers2, shuffle=True, random_state=42)
    Y1 = np.random.binomial(n=1,p=ratio,size=[n_samples//2])
    Y2 = np.random.binomial(n=1,p=ratio,size=[n_samples//2])
    Y1 = np.concatenate((Y1[:(n_samples//2)],arrRatioMaintain1(Y1[(n_samples//2)::],ratio)),axis = 0)
    Y2 = np.concatenate((Y2[:(n_samples//2)],arrRatioMaintain2(Y1[(n_samples//2)::],ratio)), axis = 0)
    X = np.concatenate((X1,X2), axis = 0 )
    Y = np.concatenate((Y1,Y2), axis = 0 )
    y = np.concatenate((Y[:(n_samples//2)],constadd(Y[(n_samples//2)::],val)),axis = 0)
    return (X,y)

if __name__ == '__main__':
    X,y = make_blobs()
    print(X.shape)
    print(y.shape)
    print(X)
    print(y)